# tm.py
#    Topic modeling with LDA, NMF, and BERTopic for a prescribed range of Scripture or other text
#    By Johnny Cheng
#    Created: 20 Jan. 2024
#    Updated: 21 Jan. 2024
#       Remove customized Chinese word embeddings


## Dependencies

import warnings
warnings.filterwarnings("ignore")

import os
import numpy as np
import pandas as pd
import string
import re

import time
from pprint import pprint
from IPython.display import IFrame
from importlib_resources import files

import jieba
from sklearn.feature_extraction.text import ENGLISH_STOP_WORDS
from nltk.stem.snowball import SnowballStemmer
from nltk.tokenize import TweetTokenizer
import nltk

from gensim.models import Word2Vec
from gensim import corpora, models
from gensim.models.coherencemodel import CoherenceModel

import torch
from bertopic import BERTopic
from transformers import BertTokenizer, BertModel


## LDA Class
class LDA:
    num_topics = 10

    def __init__(self, textfile, chi):
        self.textfile = textfile
        self.chi = chi
        self.docs = None
        self.pro_docs = None
        self.dictionary = None
        self.corpus = None
        self.model = None
        self.vis_data = None
        self.vis_file = None

    def load_text(self):
        df = pd.read_csv(self.textfile)
        self.docs = list(df.text)
        return self.docs
 
    def load_bible(self):
        # self.textfile = "web.csv"
        scfile = files('wordtm.data').joinpath(self.textfile)
        print("Loading Bible '%s' ..." %scfile)
        df = pd.read_csv(scfile)

        # Group verses into chapters
        df = df.groupby(['book_no', 'chapter'])\
                        .agg({'text': lambda x: ' '.join(x)})\
                .reset_index()

        df.text = df.text.str.replace('　', '')
        self.docs = list(df.text)
        return self.docs
   
    def process_text(self, doc):
        # List of punctuation
        punc = list(set(string.punctuation))

        # List of stop words
        add_stop = []
        stop_words = ENGLISH_STOP_WORDS.union(add_stop)

        doc = TweetTokenizer().tokenize(doc)
        doc = [each.lower() for each in doc]
        doc = [re.sub('[0-9]+', '', each) for each in doc]
        doc = [SnowballStemmer('english').stem(each) for each in doc]
        doc = [w for w in doc if w not in punc]
        doc = [w for w in doc if w not in stop_words]
        return doc
    
    def preprocess(self):
        self.pro_docs = [self.process_text(doc) for doc in self.docs]

        # Create a dictionary and corpus for LDA
        self.dictionary = corpora.Dictionary(self.pro_docs)
        self.corpus = [self.dictionary.doc2bow(doc) for doc in self.pro_docs]

    def preprocess_chi(self):
        # Build stop words
        stop_file = files('wordtm.data').joinpath("tc_stopwords_2.txt")
        stopwords = [k[:-1] for k in open(stop_file, encoding='utf-8')\
                     .readlines() if k != '']

        # Tokenize"the text using Jieba
        dict_file = files('wordtm.data').joinpath("user_dict_4.txt")
        jieba.load_userdict(str(dict_file))
        docs = [jieba.cut(doc) for doc in self.docs]

        # Replace special characters
        docs = [[word.replace('\u3000', ' ') for word in doc] \
                                     for doc in docs]

        # Remove stop words
        self.pro_docs = [' '.join([word for word in doc if word not in stopwords]) \
                                        for doc in docs]

        self.pro_docs = [doc.split() for doc in self.pro_docs]

        # Create a dictionary and corpus
        self.dictionary = corpora.Dictionary(self.pro_docs)
        self.corpus = [self.dictionary.doc2bow(doc) for doc in self.pro_docs]


    # Train LDA model
    def fit(self):
        self.model = models.LdaModel(self.corpus, 
                            num_topics=self.num_topics, 
                            id2word=self.dictionary, 
                            passes=10)
    
    def viz(self):
        self.vis_data = gensimvis.prepare(self.model, self.corpus, self.dictionary)
        pyLDAvis.enable_notebook()
        pyLDAvis.display(self.vis_data)
        print("If no visualization is shown,")
        print("  you may execute 'pyLDAvis.display(lda.vis_data)'")
        print("  to show the visualization.")

    def show_topics(self):
        print("\nTopics from LDA Model:")
        pprint(self.model.print_topics())
    
    def evaluate(self):
        # Compute coherence score
        coherence_model = CoherenceModel(model=self.model,
                                         texts=self.pro_docs,
                                         dictionary=self.dictionary,
                                         coherence='c_v')
        print(f"  Coherence: {coherence_model.get_coherence()}")
        
        # Compute perplexity
        perplexity = self.model.log_perplexity(self.corpus)
        print(f"  Perplexity: {perplexity}")
        
        # Compute topic diversity
        topic_sizes = [len(self.model[self.corpus[i]]) for i in range(len(self.corpus))]
        total_docs = sum(topic_sizes)
        topic_diversity = sum([(size/total_docs)**2 for size in topic_sizes])
        print(f"  Topic diversity: {topic_diversity}")
        
        # Compute topic size distribution
        # topic_sizes = [len(self.model[self.corpus[i]]) for i in range(len(self.corpus))]
        topic_size_distribution = max(topic_sizes) / sum(topic_sizes)
        print(f"  Topic size distribution: {topic_size_distribution}")

## End of LDA Class

## LDA Process
def lda_process(doc_file, chi=False, eval=False):
    lda = LDA(doc_file, chi)
    lda.load_bible()
    print("Bible loaded!")

    if chi:
        lda.preprocess_chi()
    else:
        lda.preprocess()
    print("Text preprocessed!")

    lda.fit()
    print("Text trained!")
    lda.viz()
    print("Visualization prepared!")
    lda.show_topics()

    if eval:
        print("\nModel Evaluation Scores:")
        lda.evaluate()
    return lda


## NMF Class
class NMF:
    num_topics = 10

    # List of punctuation
    punc = list(set(string.punctuation))

    # List of stop words
    add_stop = []
    stop_words = ENGLISH_STOP_WORDS.union(add_stop)

    def __init__(self, textfile, chi=False):
        self.textfile = textfile
        self.chi = chi
        self.docs = None
        self.pro_docs = None
        self.dictionary = None
        self.corpus = None
        self.model = None
        self.vis_data = None
        self.vis_file = None

    def load_text(self):
        df = pd.read_csv(self.textfile)
        self.docs = list(df.text)
        return self.docs
 
    def load_bible(self):
        # self.textfile = "web.csv"
        scfile = files('wordtm.data').joinpath(self.textfile)
        print("Loading Bible '%s' ..." %scfile)
        df = pd.read_csv(scfile)

        # Group verses into chapters
        df = df.groupby(['book_no', 'chapter'])\
                        .agg({'text': lambda x: ' '.join(x)})\
                .reset_index()

        df.text = df.text.str.replace('　', '')
        self.docs = list(df.text)
        return self.docs
   
    def process_text(self, doc):
        doc = TweetTokenizer().tokenize(doc)
        doc = [each.lower() for each in doc]
        doc = [re.sub('[0-9]+', '', each) for each in doc]
        doc = [SnowballStemmer('english').stem(each) for each in doc]
        doc = [w for w in doc if w not in self.punc]
        doc = [w for w in doc if w not in self.stop_words]
        return doc
    
    def preprocess(self):
        self.pro_docs = [self.process_text(doc) for doc in self.docs]

        # Create a dictionary and corpus for NMF
        self.dictionary = corpora.Dictionary(self.pro_docs)
        self.corpus = [self.dictionary.doc2bow(doc) for doc in self.pro_docs]

    def preprocess_chi(self):
        # Build stop words
        stop_file = files('wordtm.data').joinpath("tc_stopwords_2.txt")
        stopwords = [k[:-1] for k in open(stop_file, encoding='utf-8')\
                     .readlines() if k != '']

        # Tokenize"the text using Jieba
        dict_file = files('wordtm.data').joinpath("user_dict_4.txt")
        jieba.load_userdict(dict_file)
        docs = [jieba.cut(doc) for doc in self.docs]

        # Replace special characters
        docs = [[word.replace('\u3000', ' ') for word in doc] \
                                     for doc in docs]

        # Remove stop words
        self.pro_docs = [' '.join([word for word in doc if word not in stopwords]) \
                                        for doc in docs]

        self.pro_docs = [doc.split() for doc in self.pro_docs]

        # Create a dictionary and corpus
        self.dictionary = corpora.Dictionary(self.pro_docs)
        self.corpus = [self.dictionary.doc2bow(doc) for doc in self.pro_docs]


    # Train NMF model
    def fit(self):
        self.model = models.Nmf(self.corpus,
                                num_topics=self.num_topics)

    def show_topics_words(self):
        print("\nTopics-Words from NMF Model:")
        for topic_id in range(self.model.num_topics):
            topic_words = self.model.show_topic(topic_id, topn=10)
            print(f"Topic {topic_id+1}:")
            for word_id, prob in topic_words:
                # word = self.dictionary.id2token[int(word_id)]
                word = self.dictionary[int(word_id)]
                print("%s (%.6f)" %(word, prob))
            print()

    def evaluate(self):
        # Compute coherence score
        coherence_model = CoherenceModel(model=self.model,
                                         texts=self.pro_docs,
                                         dictionary=self.dictionary,
                                         coherence='c_v')
        print(f"  Coherence: {coherence_model.get_coherence()}")
        
        # Compute topic diversity
        topic_sizes = [len(self.model[self.corpus[i]]) for i in range(len(self.corpus))]
        total_docs = sum(topic_sizes)
        topic_diversity = sum([(size/total_docs)**2 for size in topic_sizes])
        print(f"  Topic diversity: {topic_diversity}")
        
        # Compute topic size distribution
        # topic_sizes = [len(self.model[self.corpus[i]]) for i in range(len(self.corpus))]
        topic_size_distribution = max(topic_sizes) / sum(topic_sizes)
        print(f"  Topic size distribution: {topic_size_distribution}")

## End of NMF Class

## NMF Process
def nmf_process(doc_file, chi=False, eval=False):
    nmf = NMF(doc_file, chi)
    nmf.load_bible()
    print("Bible loaded!")

    if chi:
        nmf.preprocess_chi()
    else:
        nmf.preprocess()
    print("Text preprocessed!")

    nmf.fit()
    print("Text trained!")
    nmf.show_topics_words()

    if eval:
        print("\nModel Evaluation Scores:")
        nmf.evaluate()

    return nmf


## BTM Class
class BTM:
    num_topics = 10

    def __init__(self, textfile, chi=False, embed=True):
        self.textfile = textfile
        self.chi = chi
        self.docs = None
        self.pro_docs = None
        self.dictionary = None
        self.corpus = None
        self.model = None
        self.vis_data = None
        self.vis_file = None

        self.embed = embed
        self.btokenizer = None # Tokenizer from pretrained model
        self.bmodel = None # BERT pretrained model
        self.bt_topics = None
        self.bt_vectorizer = None
        self.bt_analyzer = None
        self.cleaned_docs = None

    def load_text(self):
        df = pd.read_csv(self.textfile)
        self.docs = list(df.text)
        return self.docs
 
    def load_bible(self):
        # self.textfile = "web.csv"
        scfile = files('wordtm.data').joinpath(self.textfile)
        print("Loading Bible '%s' ..." %scfile)
        df = pd.read_csv(scfile)

        # Group verses into chapters
        df = df.groupby(['book_no', 'chapter'])\
                        .agg({'text': lambda x: ' '.join(x)})\
                .reset_index()

        df.text = df.text.str.replace('　', '')
        self.docs = list(df.text)
        return self.docs
   
    def process_text(self, doc):
        # List of punctuation
        punc = list(set(string.punctuation))

        # List of stop words
        add_stop = []
        stop_words = ENGLISH_STOP_WORDS.union(add_stop)

        doc = TweetTokenizer().tokenize(doc)
        doc = [each.lower() for each in doc]
        doc = [re.sub('[0-9]+', '', each) for each in doc]
        doc = [SnowballStemmer('english').stem(each) for each in doc]
        doc = [w for w in doc if w not in punc]
        doc = [w for w in doc if w not in stop_words]
        return doc
    
    def preprocess(self):
        self.pro_docs = [self.process_text(doc) for doc in self.docs]

        # Create a dictionary and corpus
        self.dictionary = corpora.Dictionary(self.pro_docs)
        self.corpus = [self.dictionary.doc2bow(doc) for doc in self.pro_docs]

    def preprocess_chi(self):
        # Build stop words
        stop_file = files('wordtm.data').joinpath("tc_stopwords_2.txt")
        stopwords = [k[:-1] for k in open(stop_file, encoding='utf-8')\
                     .readlines() if k != '']

        # Tokenize"the text using Jieba
        dict_file = files('wordtm.data').joinpath("user_dict_4.txt")
        jieba.load_userdict(dict_file)
        docs = [jieba.cut(doc) for doc in self.docs]

        # Replace special characters
        docs = [[word.replace('\u3000', ' ') for word in doc] \
                                     for doc in docs]

        # Remove stop words
        self.pro_docs = [' '.join([word for word in doc if word not in stopwords]) \
                                        for doc in docs]

        self.pro_docs = [doc.split() for doc in self.pro_docs]

        # Create a dictionary and corpus
        self.dictionary = corpora.Dictionary(self.pro_docs)
        self.corpus = [self.dictionary.doc2bow(doc) for doc in self.pro_docs]


    # Train BERTopic model - English
    def fit(self):
        j_pro_docs = [" ".join(doc) for doc in self.pro_docs]

        if self.embed:
            self.bmodel = BertModel.from_pretrained('bert-base-uncased')
            self.model = BERTopic(language='english', 
                                  calculate_probabilities=True,
                                  embedding_model=self.bmodel)
        else:
            self.model = BERTopic(language='english', 
                                  calculate_probabilities=True)

        self.bt_topics, _ = self.model.fit_transform(j_pro_docs)


    # Train BERTopic model - Chinese
    def fit_chi(self):
        j_pro_docs = [" ".join(doc) for doc in self.pro_docs]

        if self.embed:
            self.bmodel = BertModel.from_pretrained('bert-base-chinese')
            self.model = BERTopic(language='chinese (traditional)', 
                                  calculate_probabilities=True,
                                  embedding_model=self.bmodel)
        else:
            self.model = BERTopic(language='chinese (traditional)', 
                                  calculate_probabilities=True)

        self.bt_topics, _ = self.model.fit_transform(j_pro_docs)


    def show_topics(self):
        print("\nTopics from BERTopic Model:")

        for topic, words in self.model.get_topic_freq().head(self.num_topics).iterrows():
            if not self.model.get_topic(topic): continue
            twords = [word for (word, _) in self.model.get_topic(topic)]
            print(f"Topic {topic}: {' | '.join(twords)}")


    def pre_evaluate(self):
        doc_df = pd.DataFrame({"Document": self.docs,
                       "ID": range(len(self.docs)),
                       "Topic": self.bt_topics})
        documents_per_topic = doc_df.groupby(['Topic'], \
                             as_index=False).agg({'Document': ' '.join})
        self.cleaned_docs = self.model._preprocess_text(\
                              documents_per_topic.Document.values)

        # Extract vectorizer and analyzer from BERTopic
        self.bt_vectorizer = self.model.vectorizer_model
        self.bt_analyzer = self.bt_vectorizer.build_analyzer()

    def evaluate(self):
        self.pre_evaluate()
        
        # Extract features for Topic Coherence evaluation
        # words = self.bt_vectorizer.get_feature_names_out()
        tokens = [self.bt_analyzer(doc) for doc in self.cleaned_docs]

        self.dictionary = corpora.Dictionary(tokens)
        self.corpus = [self.dictionary.doc2bow(doc) for doc in tokens]

        topic_words = [[words for words, _ in self.model.get_topic(topic)] 
                       for topic in range(len(set(self.bt_topics))-1)]

        coherence = CoherenceModel(topics=topic_words, texts=tokens, corpus=self.corpus, 
                                      dictionary=self.dictionary, coherence='c_v')\
                        .get_coherence()

        print(f"  Coherence: {coherence}")


    def viz(self):
        print("\nBERTopic Model Visualization:")

        # Intertopic Distance Map
        self.model.visualize_topics().show()

        # Visualize Terms (Topic Word Scores)
        self.model.visualize_barchart().show()

        # Visualize Topic Similarity
        self.model.visualize_heatmap().show()
        
        print("  If no visualization is shown,")
        print("    you may execute the following commands one-by-one:")
        print("      btm.model.visualize_topics()")
        print("      btm.model.visualize_barchart()")
        print("      btm.model.visualize_heatmap()")
        print()

## End of BTM Class

## BTM Process
def btm_process(doc_file, chi=False, eval=False):
    btm = BTM(doc_file, chi)
    btm.load_bible()
    print("Bible loaded!")

    if chi:
        btm.preprocess_chi()
        print("Chinese text preprocessed!")
        btm.fit_chi()
    else:
        btm.preprocess()
        print("Text preprocessed!")
        btm.fit()
    
    print("Text trained!")

    btm.show_topics()

    if eval:
        print("\nModel Evaluation Scores:")
        btm.evaluate()
    
    btm.viz()

    return btm

### End of TM Module


# quot.py
#    Locate the Scripture in OT quoted from an NT verse
#    By Johnny Cheng
#    Updated: 18 Dec. 2023


import warnings
warnings.filterwarnings("ignore")

import re
import jieba

from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.metrics.pairwise import cosine_similarity

from wordtm import util


chinese_stopwords = ["、","。","〈","〉","《","》","一","一些","一何","一切","一則","一方面","一旦","一來","一樣","一般","一轉眼","七","萬一","三","上","上下","下","不","不僅","不但","不光","不單","不只","不外乎","不如","不妨","不盡","不儘然","不得","不怕","不惟","不成","不拘","不料","不是","不比","不然","不特","不獨","不管","不至於","不若","不論","不過","不問","與","與其","與其說","與否","與此同時","且","且不說","且說","兩者","個","個別","中","臨","為","為了","為什麼","為何","為止","為此","為著","乃","乃至","乃至於","麼","之","之一","之所以","之類","烏乎","乎","乘","九","也","也好","也罷","了","二","二來","於","於是","於是乎","云云","雲爾","五","些","亦","人","人們","人家","什","什麼","什麼樣","今","介於","仍","仍舊","從","從此","從而","他","他人","他們","他們們","以","以上","以為","以便","以免","以及","以故","以期","以來","以至","以至於","以致","們","任","任何","任憑","會","似的","但","但凡","但是","何","何以","何況","何處","何時","餘外","作為","你","你們","使","使得","例如","依","依據","依照","便於","俺","俺們","倘","倘使","倘或","倘然","倘若","借","借儻然","假使","假如","假若","做","像","兒","先不先","光是","全體","全部","八","六","兮","共","關於","關於具體地說","其","其一","其中","其二","其他","其餘","其它","其次","具體地說","具體說來","兼之","內","再","再其次","再則","再有","再者","再者說","再說","冒","沖","況且","幾","幾時","凡","凡是","憑","憑藉","出於","出來","分","分別","則","則甚","別","別人","別處","別是","別的","別管","別說","到","前後","前此","前者","加之","加以","即","即令","即使","即便","即如","即或","即若","卻","去","又","又及","及","及其","及至","反之","反而","反過來","反過來說","受到","另","另一方面","另外","另悉","只","只當","只怕","只是","只有","只消","只要","只限","叫","叮咚","可","可以","可是","可見","各","各個","各位","各種","各自","同","同時","後","後者","向","向使","向著","嚇","嗎","否則","吧","吧噠","含","吱","呀","呃","嘔","唄","嗚","嗚呼","呢","呵","呵呵","呸","呼哧","咋","和","咚","咦","咧","咱","咱們","咳","哇","哈","哈哈","哉","哎","哎呀","哎喲","嘩","喲","哦","哩","哪","哪個","哪些","哪兒","哪天","哪年","哪怕","哪樣","哪邊","哪裡","哼","哼唷","唉","唯有","啊","啐","啥","啦","啪達","啷當","喂","喏","喔唷","嘍","嗡","嗡嗡","呵","嗯","噯","嘎","嘎登","噓","嘛","嘻","嘿","嘿嘿","四","因","因為","因了","因此","因著","因而","固然","在","在下","在於","地","基於","處在","多","多麼","多少","大","大家","她","她們","好","如","如上","如上所述","如下","如何","如其","如同","如是","如果","如此","如若","始而","孰料","孰知","甯","寧可","寧願","寧肯","它","它們","對","對於","對待","對方","對比","將","小","爾","爾後","爾爾","尚且","就","就是","就是了","就是說","就算","就要","盡","儘管","儘管如此","豈但","己","已","已矣","巴","巴巴","年","並","並且","庶乎","庶幾","開外","開始","歸","歸齊","當","當地","當然","當著","彼","彼時","彼此","往","待","很","得","得了","怎","怎麼","怎麼辦","怎麼樣","怎奈","怎樣","總之","總的來看","總的來說","總的說來","總而言之","恰恰相反","您","惟其","慢說","我","我們","或","或則","或是","或曰","或者","截至","所","所以","所在","所幸","所有","才","才能","打","打從","把","抑或","拿","按","按照","換句話說","換言之","據","據此","接著","故","故此","故而","旁人","無","無寧","無論","既","既往","既是","既然","日","時","時候","是","是以","是的","更","曾","替","替代","最","月","有","有些","有關","有及","有時","有的","望","朝","朝著","本","本人","本地","本著","本身","來","來著","來自","來說","極了","果然","果真","某","某個","某些","某某","根據","歟","正值","正如","正巧","正是","此","此地","此處","此外","此時","此次","此間","毋寧","每","每當","比","比及","比如","比方","沒奈何","沿","沿著","漫說","焉","然則","然後","然而","照","照著","猶且","猶自","甚且","甚麼","甚或","甚而","甚至","甚至於","用","用來","由","由於","由是","由此","由此可見","的","的確","的話","直到","相對而言","省得","看","眨眼","著","著呢","矣","矣乎","矣哉","離","秒","竟而","第","等","等到","等等","簡言之","管","類如","緊接著","縱","縱令","縱使","縱然","經","經過","結果","給","繼之","繼後","繼而","綜上所述","罷了","者","而","而且","而況","而後","而外","而已","而是","而言","能","能否","騰","自","自個兒","自從","自各兒","自後","自家","自己","自打","自身","至","至於","至今","至若","致","般的","若","若夫","若是","若果","若非","莫不然","莫如","莫若","雖","雖則","雖然","雖說","被","要","要不","要不是","要不然","要麼","要是","譬喻","譬如","讓","許多","論","設使","設或","設若","誠如","誠然","該","說","說來","請","諸","諸位","諸如","誰","誰人","誰料","誰知","賊死","賴以","趕","起","起見","趁","趁著","越是","距","跟","較","較之","邊","過","還","還是","還有","還要","這","這一來","這個","這麼","這麼些","這麼樣","這麼點兒","這些","這會兒","這兒","這就是說","這時","這樣","這次","這般","這邊","這裡","進而","連","連同","逐步","通過","遵循","遵照","那","那個","那麼","那麼些","那麼樣","那些","那會兒","那兒","那時","那樣","那般","那邊","那裡","都","鄙人","鑒於","針對","阿","除","除了","除外","除開","除此之外","除非","隨","隨後","隨時","隨著","難道說","零","非","非但","非徒","非特","非獨","靠","順","順著","首先","︿","！","＃","＄","％","＆","（","）","＊","＋","，","０","１","２","３","４","５","６","７","８","９","：","；","＜","＞","？","＠","［","］","｛","｜","｝","～","￥"]


def extract_quotation(text, quot_marks):
    arr = text.split(quot_marks[0])
    if len(arr) > 1:
        return arr[1].split(quot_marks[1])[0]
    else:
        return ""


def tokenize(sentence):
    without_duplicates = re.sub(r'(.)\1+', r'\1\1', sentence)
    without_punctuation = re.sub(r'[^\w]','', without_duplicates)
    # return jieba.lcut(without_duplicates)
    return jieba.cut(without_duplicates)

	
def match_text(target, sent_tokens, lang, threshold, n=5):
    result = ''
    sent_tokens.append(target)
    if lang == 'chi':
        TfidfVec = TfidfVectorizer(tokenizer=tokenize, stop_words=chinese_stopwords)
    else:
        TfidfVec = TfidfVectorizer(analyzer='word', stop_words='english')

    tfidf = TfidfVec.fit_transform(sent_tokens)
    vals = cosine_similarity(tfidf[-1], tfidf)

    results = []
    for i in range(n):
        idx = vals.argsort()[0][-2-i]
        flat = vals.flatten()
        flat.sort()
        cos_sim = flat[-2-i]
        if (cos_sim > threshold):
            results.append((cos_sim, idx))

    return results


def match_verse(i, ot_list, otdf, df, book, chap, verse, lang, threshold):
    book_cat = util.load_word('book_categories.csv')

    if lang == 'en':
        book_sname = book
        quot_marks = ['"', '"']
    else:
        book_sname = book_cat[book_cat.book_s == book].book_chi.iloc[0]
        quot_marks=['『', '』']

    averse = util.extract(df, book=book, chapter=chap, verse=verse)
    vtext = util.get_text(averse)

    quot = extract_quotation(vtext, quot_marks)
    if quot == "": return False

    nt_str = book_sname + ' ' + str(chap) + ':' + str(verse)
    print("(%2d) %-6s %s" %(i+1, nt_str, vtext))  # NT Verse

    results = match_text(quot, ot_list, lang, threshold)
    ot_list.remove(quot)

    for cos_sim, idx in results:
        sv = otdf.iloc[idx]

        if lang == 'en':
            book_sname = book
        else:
            book_sname = book_cat[book_cat.book_s == sv.book].book_chi.iloc[0]

        ot_str = book_sname + ' ' + str(sv.chapter) + ':' + str(sv.verse)
        print("     -> %.4f %-9s %s" %(cos_sim, ot_str, sv.text))  # OT Verse

    return True


def show_quot(target, source='ot', lang='en', threshold=0.5):
    util.set_lang(lang)
    ot_cat = ['tor', 'oth', 'ket', 'map', 'mip']

    if lang == 'en':
        df = util.load_word()
    else:
        df = util.load_word('cuv.csv')
        # df.text = df.text.apply(lambda x: x.replace('\u3000', ''))

    if source in ot_cat:
        otdf = util.extract(df, category=source)
    else:
        otdf = util.extract(df, testament=0)

    ot_list = util.get_text_list(otdf)

    i = 0
    for _, row in target.iterrows():
        if match_verse(i, ot_list, otdf, target, row.book, row.chapter, row.verse, lang, threshold):
            i += 1



# util.py
#   Some utility functions including loading Scriptue, setting Scripture language,
#     extracting a specifc range of Scripture
#   By Johnny Cheng
#   Updated: 22 Dec. 2023


import re
import pandas as pd
from importlib_resources import files

import nltk
from nltk.corpus import stopwords
from nltk.stem import PorterStemmer
from nltk.tokenize import word_tokenize

import jieba
from collections import Counter

chi_flag = False
glang = 'en'
stops = set()


def is_chi():
    return chi_flag


def load_text(filepath, nr=0, info=False):
    print("Loading text '%s' ..." %filepath)
    df = pd.read_csv(filepath)
    if nr > 0:
       print("Initial Records:")
       df.head(int(nr))
    if info:
        print("\nDataset Information:")
        df.info()
    return df


def load_word(ver='web.csv', nr=0, info=False):
    scfile = files('wordtm.data').joinpath(ver)
    print("Loading Scripture '%s' ..." %scfile)
    df = pd.read_csv(scfile)
    if nr > 0:
       print("Initial Records:")
       df.head(int(nr))
    if info:
        print("\nDataset Information:")
        df.info()
    return df


def group_text(df, column='chapter'):
    gdf = df.groupby(['book_no', column])\
                        .agg({'text': lambda x: ''.join(x)})\
                .reset_index()
    return gdf


def get_list(df, column='book'):
    if column in list(df.columns):
        return list(df[column].unique())
    else:
        return "No such column!"


def get_text(df):
    return ''.join(list(df.text)).replace('\u3000', '')


def get_text_list(df):
    return df.text.apply(lambda x: x.replace('\u3000', '')).tolist()


def clean_text(df):
    df.text = [re.sub(r'\d+', '', str(v).replace('\n', ' ')) for v in df.text]
    for sw in stopwords.words('english'):
        df.text = [v.replace(' ' + sw + ' ', ' ') for v in df.text]

    df.text = df.text.apply(lambda v: " ".join(w.lower() for w in v.split()))
    df.text = df.text.str.replace('[^\w\s]', '', regex=True)
    return df


## Functions for Chinese Text

def add_chi_vocab():
    vocab_file = files('wordtm.data').joinpath('bible_vocab.txt')
    print("Loading Chinese vocabulary '%s' ..." %vocab_file)
    with open(vocab_file, 'r', encoding='utf8') as f:
        vocab_list = f.readlines()
        for vocab in vocab_list:
            jieba.add_word(vocab.replace('\n', ''), freq=1000)


def chi_stops():
    dict_file = files('wordtm.dictionary').joinpath('dict.txt.big.txt')
    cloud_file = files('wordtm.dictionary').joinpath('stopWord_cloudmod.txt')
    jieba.set_dictionary(dict_file)
    with open(cloud_file, 'r', encoding='utf-8-sig') as f:
        return f.read().split('\n')


def set_lang(lang='en'):
    global glang, stops
    glang = lang
    if glang == 'en':  # English
        stops = set(stopwords.words("english"))
    else:  # Chinese
        add_chi_vocab()
        stops = chi_stops()
        chi_flag = True


def get_diction_en(df):
    words = word_tokenize(' '.join(list(df.text)))
    stem = PorterStemmer()
    
    terms = []
    for t in words:
        t = stem.stem(t)
        if t not in stops:
            terms.append(t)

    diction = Counter(terms)
    return diction


def get_diction_chi(df):
    text = ''.join(list(df.text)).replace('\u3000', '')
    text = re.sub("[、．。，！？『』「」〔〕]", "", text)

    terms = []
    for t in jieba.cut(text, cut_all=False):
        if t not in stops:
            terms.append(t)

    diction = Counter(terms)
    return diction


def get_diction(df):
    if glang == 'en':
        return get_diction_en(df)
    else:
        return get_diction_chi(df)


def chi_sent_terms(text):
    text = re.sub("[、．。，！？『』「」〔〕]", "", text)
    terms = []
    for t in jieba.cut(text, cut_all=False):
        if t not in stops:
            terms.append(t)
    return terms


def get_sent_terms(text):
    if glang == 'en':
        return word_tokenize(text)
    else:
        return chi_sent_terms(text)


## Extract Scripture by Various Levels of Filtering

# Extract Scripture by testament, category, book or chapter or verse
def extract(df, testament=-1, category='', book=0, chapter=0, verse=0):
    no_ret = "No scripture is extracted!"
    sub_df = pd.DataFrame()  # Empty DataFrame
    isbook = ischapter = False

    if (testament > -1) & (testament < 2):
        sub_df = df[df.testament==int(testament)]
    elif category != '':
        if category in get_list(df, column='category'):
            sub_df = df[df.category==category]
        elif category in get_list(df, column='cat'):
            sub_df = df[df.cat==category]
    elif book in get_list(df, column='book'):
        sub_df = df[df.book==book]
        isbook = True
    elif isinstance(book, int):
        if book > 0 & book < 67:
            sub_df = df[df.book_no==book]
            isbook = True
    elif isinstance(book, tuple):
        if (book[0] <= book[1]) & (book[0] > 0) & (book[1] < 67):
            sub_df = df[(df.book_no >= book[0]) & (df.book_no <= book[1])]
            isbook = True

    if isbook & (len(sub_df) > 0) & (chapter != 0):
        if isinstance(chapter, int):
            sub_df = sub_df[sub_df.chapter==chapter]
            ischapter = True
        elif isinstance(chapter, tuple):
            if chapter[0] <= chapter[1]:
                sub_df = sub_df[(sub_df.chapter >= chapter[0]) & (sub_df.chapter <= chapter[1])]
                ischapter = True

        if ischapter & (len(sub_df) > 0) & (verse != 0):
            if isinstance(verse, int):
                sub_df = sub_df[sub_df.verse==verse]
            elif isinstance(verse, tuple):
                if verse[0] <= verse[1]:
                    sub_df = sub_df[(sub_df.verse >= verse[0]) & (sub_df.verse <= verse[1])]

    if len(sub_df) > 0:
        return sub_df.copy()
    else:
        return no_ret


# Extract Scipture by a Filter String
def extract2(df, filter=''):
    chapter = verse = 0

    if filter == '':
        return df
    else:
        parts = filter.split()
        book = parts[0]
        if len(parts) > 1:
            parts = parts[1].split(':')
            if parts[0] == '':
                chapter = 0
            else:
                chapter = int(parts[0])

            if (len(parts) > 1):
                if (parts[1] != ''):
                    parts = parts[1].split('-')
                    if parts[0] == '':
                        verse = 1
                    else:
                        verse = int(parts[0])

                    if (len(parts) > 1):
                        if (parts[1] == ''):
                            verse = (verse, 999)
                        else:
                            verse = (verse, int(parts[1]))

        return extract(df, book=book, chapter=chapter, verse=verse)



from setuptools import setup, find_packages

VERSION = '0.3.6' 
DESCRIPTION = 'Scripture NLP Package'
LONG_DESCRIPTION = 'An NLP package for topic modeling on the Holy Scripture from low-code to pro-code'

# Setting up
setup(
        name="wordtm", 
        version=VERSION,
        author="Johnny CHENG",
        author_email="<drjohnnycheng@gmail.com>",
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        project_urls={
            'Documentation': 'https://drjohnnycheng.github.io/wordtm',
        },
        include_package_data=True,
        packages=find_packages(),
        package_data={
            'wordtm': ['docs/*.html', 'docs/*.css'],
        },
        install_requires=['numpy', 'pandas', 'importlib_resources', 'regex', 'nltk', \
                    'matplotlib', 'wordcloud', 'pillow', 'jieba', 'gensim', 'pyLDAvis',  \
                    'bertopic',  'transformers', 'spacy', 'seaborn', \
                    'importlib', 'networkx', 'plotly', 'IPython', 'scikit-learn', 'torch'],
        
        keywords=['word', 'scripture', 'topic modeling', 'visualization', \
                  'low-code', 'pro-code', 'network analysis', 'BERTopic', \
                  'LDA', 'NFM'],

        classifiers= [
            "Development Status :: 3 - Alpha",
            "Intended Audience :: Education",
            "Intended Audience :: Science/Research",
            "Intended Audience :: Religion",
            "Programming Language :: Python :: 3",
            "Operating System :: Microsoft :: Windows",
            "Operating System :: MacOS :: MacOS X",
        ]
)
